package xyz.matj1.bp;

import mcdr.objects.impl.ObjectMotionWord;
import mcdr.sequence.SequenceMotionWords;
import mcdr.test.utils.ObjectCategoryMgmt.Category;
import messif.objects.LocalAbstractObject;
import org.jfree.chart.ChartTheme;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.util.ExportUtils;
import xyz.matj1.bp.experiments.KnnPrecision;
import xyz.matj1.bp.themes.XyGraphTheme;
import xyz.matj1.bp.types.Vector2;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class Utils {
	/**
	 * Returns a string representing the given ObjectMotionWord.
	 *
	 * @param seq the sequence castable to ObjectMotionWord
	 * @return a string representing the input sequence
	 */
	public static String sequenceToString(LocalAbstractObject seq) {
		StringBuilder rtv = new StringBuilder();
		for (ObjectMotionWord mw : ((SequenceMotionWords<? extends ObjectMotionWord>)seq).getSequenceData()) {
			if (rtv.length() > 0) {
				rtv.append(',');
			}
			rtv.append(mw.toString());
		}
		return seq.getLocatorURI() + " [" + rtv + "]";
	}

	/**
	 * Calculates the distance of two categories as the average
	 * of the distances of objects in each tuples from the Cartesian product of given collections.
	 *
	 * @param objectsInCategory1 a collection of the objects in the first category
	 * @param objectsInCategory2 a collection of the objects in the second category
	 * @return the distance of of the categories
	 */
	public static Double distanceOfCategories(
		Collection<LocalAbstractObject> objectsInCategory1,
		Collection<LocalAbstractObject> objectsInCategory2
	) {
		ArrayList<Float> objectDistances = new ArrayList<>();
		for (LocalAbstractObject object1 : objectsInCategory1) {
			for (LocalAbstractObject object2 : objectsInCategory2) {
				objectDistances.add(object1.getDistance(object2));
			}
		}
		return averageOfFloats(objectDistances);
	}

	/**
	 * Returns the average of all the values in the input collection.
	 *
	 * @param collection the input collection
	 * @return average of the values in the input collection
	 */
	public static Double average(Collection<Double> collection) {
		return collection.parallelStream().mapToDouble(x -> x).average().orElse(Double.NaN);
	}

	/**
	 * Returns the average of all the values in the input collection.
	 *
	 * @param collection the input collection
	 * @return average of the values in the input collection
	 */
	public static Double averageOfFloats(Collection<Float> collection) {
		return collection.parallelStream().mapToDouble(x -> (double)x).average().orElse(Double.NaN);
	}

	/**
	 * Converts the given {@link Iterator} to {@link Stream} by putting each object returned from the iterator to the stream.
	 *
	 * @param iterator the given iterator
	 * @return a stream with the values taken from the iterator
	 */
	public static <T> Stream<T> iteratorToStream(Iterator<T> iterator) {
		Stream.Builder<T> builder = Stream.builder();
		iterator.forEachRemaining(builder::add);
		return builder.build().parallel();
	}

	/**
	 * Returns the length in poses of the given motion words sequence.
	 *
	 * @param object a motion words sequence
	 * @return the length of object
	 * @throws IllegalArgumentException if the given object isn't a motion words sequence
	 */
	public static int getLength(LocalAbstractObject object) throws IllegalArgumentException {
		try {
			return Integer.parseInt(object.getLocatorURI().split("_")[3]);
		} catch (IndexOutOfBoundsException | NumberFormatException e) {
			throw new IllegalArgumentException("The given object probably isn't a motion words sequence", e);
		}
	}

	/**
	 * This normalises a <code>value</code> in a way that,
	 * if its absolute value is smaller than 1, this returns 1/<code>value</code>,
	 * otherwise it returns <code>value</code>.
	 *
	 * @param value a value to be normalised
	 * @return <code>value</code> normalised
	 */
	public static double normaliseQuotient(double value) {
		return (Math.abs(value) < 1)
			? 1 / value
			: value;
	}

	/**
	 * This converts a Double to its string representation with <code>decimals</code> decimal places
	 * and ‘.’ as the separator of the whole part and the fractional part.
	 *
	 * @param input what shall be represented
	 * @param decimals decimal places of the representation
	 * @return a String representation of the given Double
	 */
	public static String displayDouble(Double input, int decimals) {
		double multiplier = Math.pow(10, decimals);
		return ((Double)(Math.round(input * multiplier) / multiplier)) // round to `decimals` decimal places
			.toString().replaceAll("\\.?0*$", ""); // strip maybe ‘.’ and ‘0’s from the end
	}

	/**
	 * This prints given string to given writers.
	 *
	 * @param data what shall be printed
	 * @param outputs where it shall be printed
	 * @throws IOException if writing to a given writer throws this exception
	 */
	public static void write(String data, Writer... outputs) throws IOException {
		for (var output : outputs) {
			output.write(data);
			output.flush();
		}
	}

	/**
	 * This takes the output from {@link KnnPrecision#average(Map)}
	 * or {@link StretchedSequences#knnPrecisionAverage(Map)}
	 * and formats it to a string: one entry per line; the key, a tab, and the value.
	 * The key is represented by its id number in the string representation.
	 *
	 * @param averageData output from <code>average</code> to be formatted
	 * @return a string representation of the input
	 */
	public static String knnPrecisionAverageFormat(Map<Category, Double> averageData) {
		return averageData.entrySet().parallelStream()
			.sorted(Entry.comparingByKey(Comparator.comparingInt(c -> c.index)))
			.map(entry -> String.format("%s\t%f%n", entry.getKey().id, entry.getValue()))
			.collect(Collectors.joining(System.lineSeparator(), "", System.lineSeparator()));
	}

	/**
	 * This creates a scatter-plot graph with the default theme
	 * ({@link XyGraphTheme}) and given format.
	 * The format of the output is PDF.
	 * This may change the given <code>chart</code> if the <code>chart</code> can't be easily cloned.
	 *
	 * @param chart a chart containing the data for the graph
	 * @param output where the graph shall be saved
	 * @param size the width and the height of the output (in pixels)
	 */
	public static void writeXYGraphAsPdf(
		JFreeChart chart,
		File output,
		Vector2<Double> size
	) {
		writeXYGraphAsPdf(chart, output, size, new XyGraphTheme());
	}

	/**
	 * This creates a scatter-plot graph with a given theme and format.
	 * The format of the output is PDF.
	 * This may change the given <code>chart</code> if the <code>chart</code> can't be easily cloned.
	 *
	 * @param chart a chart containing the data for the graph
	 * @param output where the graph shall be saved
	 * @param size the width and the height of the output (in pixels)
	 */
	public static void writeXYGraphAsPdf(
		JFreeChart chart,
		File output,
		Vector2<Double> size,
		ChartTheme theme
	) {
		try {
			chart = (JFreeChart)chart.clone();
		} catch (CloneNotSupportedException | ClassCastException ignored) {}

		theme.apply(chart);

		assert ExportUtils.isOrsonPDFAvailable();
		ExportUtils.writeAsPDF(chart, size.a().intValue(), size.b().intValue(), output);
	}

	/**
	 * This superimposes all datasets from the charts in <code>charts</code> to a single chart.
	 * So this adds all datasets from XYPlots of the charts to a single chart and returns the chart.
	 *
	 * The returned chart is a clone of the first chart from the iterator of <code>charts</code>.
	 * But, if it doesn't support cloning, the returned chart is the first chart from the iterator of <code>charts</code>.
	 * That means that the chart is changed
	 *
	 * @param charts a nonempty collection of charts
	 * @return a clone (or maybe original) of the first chart with added datasets of all other charts
	 * @throws IllegalArgumentException if <code>charts</code> is empty
	 */
	public static JFreeChart superimpose(Collection<JFreeChart> charts) {
		if (charts.isEmpty()) {
			throw new IllegalArgumentException("The collection of charts is empty.");
		}

		var chartIterator = charts.iterator();
		var firstChart = chartIterator.next();

		JFreeChart tempNewChart;
		try {
			tempNewChart = (JFreeChart)firstChart.clone();
		} catch (CloneNotSupportedException e) {
			tempNewChart = firstChart;
		}
		final var newChart = tempNewChart;

		var newPlot = newChart.getXYPlot();

		chartIterator.forEachRemaining(chart -> {
			var plot = chart.getXYPlot();
			for (int i = 0; i < plot.getDatasetCount(); i++) {
				newPlot.setDataset(newPlot.getDatasetCount(), plot.getDataset(i));
		}});

		return newChart;
	}
}
