package xyz.matj1.bp.experiments;

import mcdr.objects.impl.ObjectMotionWordNMatches;
import mcdr.test.utils.ObjectCategoryMgmt.Category;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.DefaultXYDataset;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import xyz.matj1.bp.Utils;
import xyz.matj1.bp.enums.LengthRelation;
import xyz.matj1.bp.types.DoubleRange;
import xyz.matj1.bp.types.Vector2;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.function.BiFunction;
import java.util.stream.Collectors;

/**
 * This is for functions used for the experiment pairwise correlation.
 */
public class PairwiseCorrelation {
	/**
	 * This calculates the length relation and the distance for each pair of sequences in each category.
	 *
	 * @param mwMgmt ObjectMgmt object containing the sequences and their categories
	 * @param lengthRelation how the relation of the lengths shall be expressed
	 * @return a map of categories to maps of pairs of the sequences to the results
	 */
	public static PairwiseCorrelationData compute(
		ObjectMgmt mwMgmt,
		LengthRelation lengthRelation
	) {
		var categoryObjects = mwMgmt.getCategoryObjects();
		var accumulator = new PairwiseCorrelationData(LengthRelation.QUOTIENT);

		for (var entry : categoryObjects.entrySet()) {
			var category = entry.getKey();
			var sequences = entry.getValue();
			accumulator.put(category, new ConcurrentHashMap<>());

			for (int i1 = 0; i1 < sequences.size(); i1++) {
				for (int i2 = i1+1; i2 < sequences.size(); i2++) {
					var seq1 = sequences.get(i1);
					var seq2 = sequences.get(i2);

					if (category.id.equals("73") && Utils.normaliseQuotient((double)Utils.getLength(seq1) / Utils.getLength(seq2)) > 10) {
						continue;
					} // to skip anomalous data points caused by an error in the data set

					BiFunction<Double, Double, Double> relationOperation = switch (lengthRelation) {
						case DIFFERENCE -> (a, b) -> Math.abs(a - b);
						case QUOTIENT -> (a, b) -> Utils.normaliseQuotient(a / b);
					};
					accumulator.get(category).put(
						new Vector2<>(seq1, seq2),
						new Vector2<>(
							relationOperation.apply((double)Utils.getLength(seq1), (double)Utils.getLength(seq2)),
							(double)seq1.getDistance(seq2)
					));
		}}}

		return accumulator;
	}

	/**
	 * This calls {@link #show(ObjectMgmt, LengthRelation, File, Vector2)} with a default size of output 1024×1024.
	 *
	 * @param mwMgmt ObjectMgmt object containing the sequences and their categories
	 * @param lengthRelation how the relation of the lengths shall be expressed
	 * @param graphsPath path of the directory where the graphs shall be saved
	 * @throws IOException if there is an error in writing to a file in graphsPath
	 */
	public static void show(
		ObjectMgmt mwMgmt,
		LengthRelation lengthRelation,
		File graphsPath
	) throws IOException {
		var defaultSize = new Vector2<>(1024d, 1024d);

		show(
			mwMgmt,
			lengthRelation,
			graphsPath,
			defaultSize
	);}

	/**
	 * Creates an XY graph of the length relation and the distance for each pair of sequences in the category for each category.
	 * That means that the points of each category are shown in their own graph, separated from the points of other categories.
	 *
	 * @param mwMgmt ObjectMgmt object containing the sequences and their categories
	 * @param lengthRelation how the relation of the lengths shall be expressed
	 * @param graphsPath path of the directory where the graphs shall be saved
	 * @param size a pair of integers specifying the width and the height of the output image
	 * @throws IOException if there is an error in writing to a file in graphsPath
	 */
	public static void show(
		ObjectMgmt mwMgmt,
		LengthRelation lengthRelation,
		File graphsPath,
		Vector2<Double> size
	) throws IOException {
		ObjectMotionWordNMatches.maxPartsToMatch = 5;
		ObjectMotionWordNMatches.nMatches = 1; //another reasonable choice is 2

		if (!graphsPath.isDirectory()) {
			throw new IllegalArgumentException(graphsPath.getPath() + " isn't a directory.");
		}

		var pairwiseCorrelationMap = compute(mwMgmt, lengthRelation);

		for (var entry : pairwiseCorrelationMap.entrySet()) {
			var category = entry.getKey();
			var chart = showCategory(entry.getValue().values(), lengthRelation, Optional.of(category));
			Utils.writeXYGraphAsPdf(chart, new File(graphsPath, "graph" + category.id + ".pdf"), size);
		}

		System.err.println("Pairwise correlations for " + mwMgmt + " written to " + graphsPath.getPath());
	}

	/**
	 * This creates an XY graph of pairwise correlations treating them as in a single category.
	 * Iff <code>categoryOpt</code> is empty, title is null, so it needs to be added later.
	 *
	 * @param pointData
	 */
	public static JFreeChart showCategory(Collection<Vector2<Double>> pointData, LengthRelation lengthRelation, Optional<Category> categoryOpt) {
		var dataSeries = new XYSeries(categoryOpt.orElse(new Category(null, null, 0)).index);
		pointData.forEach(v -> dataSeries.add(
			switch (lengthRelation) {
				case QUOTIENT -> v.a(); //Math.log(v.a()) / Math.log(2);
				case DIFFERENCE -> v.a();
			},
			v.b()
		));
		var dataset = new XYSeriesCollection(dataSeries);

		return ChartFactory.createScatterPlot(
			categoryOpt.isEmpty() ? null : "The length relation and the distance of each two sequences in category " + categoryOpt.get().id,
			"the " + switch(lengthRelation) {
				case DIFFERENCE -> "difference";
				case QUOTIENT -> "quotient"; //"base-2 logarithm of the quotient"
			} +" of the lengths",
			"the distance",
			dataset,
			PlotOrientation.VERTICAL,
			false,
			false,
			false
		);
	}

	/**
	 * This removes all sequence identifiers from the data,
	 * puts all points from all categories together,
	 * and smooths the points by calculating an average of all points
	 *
	 * @param pairData
	 * @param precision
	 * @param averageRange
	 * @return
	 */
	public static List<Vector2<Double>> smooth(PairwiseCorrelationData pairData, Double precision, Double averageRange) {
		var flatData = new ArrayList<Vector2<Double>>(); // pairData flattenened and stripped of the sequence identifiers
		pairData.forEach(
			(category, data) -> data.forEach(
				(objectVector, point) -> flatData.add(point)
		));

		return smooth(flatData, precision, averageRange);
	}

	/**
	 * This smooths the given data by averaging all b values in the range of
	 * <code>averageRange</code> from each point.
	 * <code>precision</code> determines the frequency
	 * of the calculated average values on the a (x) axis.
	 *
	 * @param pairData the data to be smoothed
	 * @param precision the frequency of average points
	 * @param averageRange points up to how far from the centre shall be counted in the average
	 * @return smoothed data sequence
	 */
	public static List<Vector2<Double>> smooth(Collection<Vector2<Double>> pairData, Double precision, Double averageRange) {
		var sortedData = new TreeSet<Vector2<Double>>(Comparator.comparing(Vector2::a));
		sortedData.addAll(pairData);

		return new DoubleRange(
			sortedData.first().a(),
			sortedData.last().a(),
			precision
		).stream().map(
			value -> new Vector2<>(
				value,
				sortedData.subSet(
					new Vector2<>(value - averageRange, 0d),
					new Vector2<>(value + averageRange, 0d)
				).stream().mapToDouble(Vector2::b).average().orElse(Double.NaN)
			)
		).toList();
	}

	/**
	 * A wrapper for the data type used as a result of {@link #compute(ObjectMgmt, LengthRelation)}
	 * also storing the length relation of the sequences (difference or quotient)
	 *
	 * Only 2 constructors are supported, the one with no parameters and the one taking a map for initialisation.
	 */
	public static class PairwiseCorrelationData extends ConcurrentHashMap<Category, ConcurrentMap<Vector2<LocalAbstractObject>, Vector2<Double>>> {
		private final LengthRelation lengthRelation;

		public PairwiseCorrelationData(LengthRelation lengthRelation) {
			super();
			this.lengthRelation = lengthRelation;
		}
		public PairwiseCorrelationData(
			LengthRelation lengthRelation,
			Map<? extends Category, ? extends ConcurrentMap<Vector2<LocalAbstractObject>, Vector2<Double>>> initMap
		) {
			super(initMap);
			this.lengthRelation = lengthRelation;
		}

		/**
		 * @return the length relation of the sequences in these data
		 */
		public LengthRelation lengthRelation() {
			return lengthRelation;
		}

		public Collection<Vector2<Double>> stripIds(Category category) {
			return this.get(category).values();
		}

		public Map<Category, Collection<Vector2<Double>>> stripIds() {
			return keySet().stream()
				.collect(Collectors.toConcurrentMap(
					k -> k,
					this::stripIds
				));
		}

		/**
		 * This converts the object to a {@link XYDataset} for creating {@link Plot}s for {@link JFreeChart}s.
		 * Each category has a data series identified by the index of the category.
		 *
		 * @return a dataset with the data in this object
		 */
		public XYDataset toDataset() {
			var data = stripIds().entrySet().stream()
				.collect(Collectors.toMap(
					entry -> entry.getKey().index,
					entry -> entry.getValue().stream().toList()
				)); // stripIds but then replaces categories with their index numbers and converts the data series to lists

			var dataset =  new DefaultXYDataset();
			for (var category : data.keySet()) {
				double[] xArray = data.get(category).stream().mapToDouble(Vector2::a).toArray();
				double[] yArray = data.get(category).stream().mapToDouble(Vector2::b).toArray();
				dataset.addSeries(category, new double[][]{xArray, yArray});
			}

			return dataset;
		}
	}
}
